/*
 * rastertospp - Samsung SPP-2020 photo printer driver for CUPS
 *
 *    based on: CUPS-DDK examples
 *     licence: GPL
 *     authors: Tomasz Koprowski (tomek [at] koprowski.org) - original code
 *            : Phil Hallows - patch for newer cups versions
 *     version: 1.2 (2015-04-13)
 *
 */

#include <cupsfilters/driver.h>

/* --------------------------------------------------------------- */

#define WIDTH 1242                                                              
#define HEIGHT 1872                                                             
#define BUFSIZE WIDTH*3

/* Anti-clipping constants for now, the formula is: v = A*v / B + C */
#define ANTICLIP_A 3
#define ANTICLIP_B 4
#define ANTICLIP_C 32

/* PJL job header */                                                            
#define PJL_HEAD "\x1b%-12345X"\
                 "@PJL DEFAULT SERVICEDATE=20071230\x0d\x0a"\
		 "@PJL SET USERNAME=\"\"\x0d\x0a"\
		 "@PJL SET JOBNAME=\"\"\x0d\x0a"\
		 "\x1b%-12345X"                                                 
#define PJL_HEAD_LEN 96                                                         
								                                                                                    
/* SPP page (?) header */                                                       
#define SPP_PAGE_HEAD "\x1b\x53\x53\x00"\
                      "\x02\x50\x03\x1b\x53\x53\x00"\
	              "\x08\x50\x02\x03\x03\x00"\
	              "\x01\x00"\
	              "\x01"
#define SPP_PAGE_HEAD_LEN 20

/* SPP block header (for 128 line block) */                                     
#define SPP_BLOCK_HEAD_128 "\x1b\x53\x53\x00\x02\x50\x07\x00\x07\x50\x00"       
#define SPP_BLOCK_HEAD_128_LEN 11                                               
                                                                                
/* SPP block header (for 80 line block) */                                      
#define SPP_BLOCK_HEAD_80 "\x1b\x53\x53\x00\x02\x50\x07\x00\x04\x92\x00"        
#define SPP_BLOCK_HEAD_80_LEN 11                                                
                                                                                
/* SPP line header */                                                           
#define SPP_LINE_HEAD "\xff\xff\xff\xff\xff\xff\xff\xff"\
                      "\xff\xff\xff\xff\xff\xff\xff\xff"\
		      "\xff\xff"                                                
#define SPP_LINE_HEAD_LEN 18                                                    
	                                                                                    
/* SPP page (?) tail */                                                         
#define SPP_PAGE_TAIL "\x1b\x53\x53\x00\x02\x50\x0b"\
                      "\x1b\x53\x53\x00\x02\x50\x04"                            
#define SPP_PAGE_TAIL_LEN 14


/* --------------------------------------------------------------- */

int currentLine = 0;
int Page;

unsigned char buff[BUFSIZE];
unsigned char *preBuff;


/* --------------------------------------------------------------- */

void	StartPage(ppd_file_t *ppd, cups_page_header2_t *header, int job_id,
	          const char *user, const char *title, int num_options,
		  cups_option_t *options);
void	EndPage(ppd_file_t *ppd, cups_page_header2_t *header);
void	OutputLine(ppd_file_t *ppd, cups_page_header2_t *header);
int	ReadLine(cups_raster_t *ras, cups_page_header2_t *header);


/* --------------------------------------------------------------- */

void StartPage(ppd_file_t *ppd, cups_page_header2_t *header,
		int job_id, const char *user, const char *title,
		int num_options, cups_option_t *options) {

  /* Just a precaution in case someone tampers with the PPD */
  
  if(header->cupsColorSpace != CUPS_CSPACE_RGB) {
    fprintf(stderr, "ERROR: Only RGB is supported.\n");
    exit(100);
  }

  if(header->HWResolution[0] != 300 || header->HWResolution[1] != 300) {
    fprintf(stderr, "ERROR: Only 300x300 dpi is supported.\n");
    exit(100);
  }  

  if(header->cupsBitsPerColor!=8 || header->cupsBitsPerPixel!=24) {
    fprintf(stderr, "ERROR: Only 24-bit supported.\n");
    exit(100);
  }

    fprintf(stderr, "DEBUG: cupsCompression = %d\n", header->cupsCompression);


  preBuff=malloc( header->cupsBytesPerLine );

  /* print PJL header */                                                      
  fwrite(PJL_HEAD, PJL_HEAD_LEN, 1, stdout); 

  /* print SPP page header */                                                 
  fwrite(SPP_PAGE_HEAD, SPP_PAGE_HEAD_LEN, 1, stdout);

}

/* --------------------------------------------------------------- */

void EndPage(ppd_file_t *ppd, cups_page_header2_t *header) {

    /* Print empty lines until we reach the end of page */
    memset(buff, 255, BUFSIZE);
    while(currentLine<HEIGHT) {
	OutputLine(ppd, header);
    }

    /* Output page closing suffix */
    fwrite(SPP_PAGE_TAIL, SPP_PAGE_TAIL_LEN, 1, stdout);
    
    free(preBuff);
}

/* --------------------------------------------------------------- */

void OutputLine(ppd_file_t *ppd, cups_page_header2_t *header) {

    int blockNumber,i;
    unsigned char r,g,b;
    
    blockNumber = currentLine/128;
    
    /* Output block header */
    if(currentLine%128==0) {
	if(blockNumber!=14) {
	    /* Normal (full sized) block */
	    fwrite(SPP_BLOCK_HEAD_128, SPP_BLOCK_HEAD_128_LEN, 1, stdout);
	} else {
	    /* Last (smaller sized) block */
	    fwrite(SPP_BLOCK_HEAD_80, SPP_BLOCK_HEAD_80_LEN, 1, stdout);
	}
    }

    /* Anti-clip */
    if(header->cupsCompression) {
	for(i=0; i<BUFSIZE; i+=3) {                                         
    	    r=buff[i];                                                      
    	    g=buff[i+1];                                                    
    	    b=buff[i+2];                                                    

	    r= ANTICLIP_A * r / ANTICLIP_B + ANTICLIP_C;
	    g= ANTICLIP_A * g / ANTICLIP_B + ANTICLIP_C;
	    b= ANTICLIP_A * b / ANTICLIP_B + ANTICLIP_C; 
			                                                                                    
    	    buff[i]=r;                                                       
    	    buff[i+1]=g;                                                    
	    buff[i+2]=b;                                                    
	}
    }
    /* write line start */                                              
    fwrite(SPP_LINE_HEAD, SPP_LINE_HEAD_LEN, 1, stdout);                
		                                                                                
    /* write line data */                                               
    fwrite(buff, BUFSIZE, 1, stdout);
    
    currentLine++;
}


/* --------------------------------------------------------------- */

int ReadLine(cups_raster_t *ras, cups_page_header2_t *header) {

    int hOff;
    int imgWidth;
    
    /* Read cups raster data */
    memset(buff, 255, BUFSIZE);    
    cupsRasterReadPixels(ras, preBuff, header->cupsBytesPerLine);

    /* Center image horizontally */
    imgWidth = header->cupsWidth;
    if(imgWidth<WIDTH) {
	hOff = 3* ((WIDTH-imgWidth)/2); 
    } else {
	hOff=0;
    }

    /* Copy to printer-line sized buffer */
    memcpy(buff+hOff, preBuff, min(BUFSIZE-hOff, header->cupsBytesPerLine));

    return 1;
}

/* --------------------------------------------------------------- */


int main(int argc, char *argv[]) {

    int			fd;		/* File descriptor */
    cups_raster_t	*ras;		/* Raster stream for printing */
    cups_page_header2_t	header;		/* Page header from file */
    int			y;		/* Current line */
    ppd_file_t		*ppd;		/* PPD file */
    //int			job_id;		/* Job ID */
    int			num_options;	/* Number of options */
    cups_option_t	*options;	/* Options */


    /* Standard CUPS rasterto* stuff follows */

    setbuf(stderr, NULL);
    if (argc < 6 || argc > 7) {
	fputs("ERROR: rastertospp job-id user title copies options [file]\n", stderr);
	return (1);
    }

    num_options = cupsParseOptions(argv[5], 0, &options);
    ppd = ppdOpenFile(getenv("PPD"));

    if (!ppd) {
	fputs("ERROR: Unable to open PPD file!\n", stderr);
	return (1);
    }

    ppdMarkDefaults(ppd);
    cupsMarkOptions(ppd, num_options, options);

    if (argc == 7) {
	if ((fd = open(argv[6], O_RDONLY)) == -1) {
	    perror("ERROR: Unable to open raster file - ");
	    return (1);
	}
    } else {
	fd = 0;
    }

    ras = cupsRasterOpen(fd, CUPS_RASTER_READ);

    //job_id = atoi(argv[1]);
    Page = 0;


    /* Print all the pages */
    while (cupsRasterReadHeader2(ras, &header)) {
	Page ++;

	fprintf(stderr, "PAGE: %d %d\n", Page, header.NumCopies);
	fprintf(stderr, "INFO: Starting page %d...\n", Page);

	StartPage(ppd, &header, atoi(argv[1]), argv[2], argv[3],
		    num_options, options);

	/* Print all the lines within the page */
	for (y = 0; y < (int)header.cupsHeight ; y ++) {
	
	    /* Progress */
	    if ((y & 127) == 0) {
		fprintf(stderr, "INFO: Printing page %d, %d%% complete...\n", Page,
	        100 * y / header.cupsHeight);
	    }
	    
	    /* Read all the pages from input but print only
	       as much as will fit on the page */
	       
	    if (ReadLine(ras, &header) && y<HEIGHT) {
		OutputLine(ppd, &header);
	    }
	}

	fprintf(stderr, "INFO: Finished page %d...\n", Page);

	EndPage(ppd, &header);
    }


    /* Close and free all the CUPS related stuff */
    cupsFreeOptions(num_options, options);
    cupsRasterClose(ras);
    if (fd != 0) {
	close(fd);
    }
    
    /* Return and print status */
    if (Page == 0) {
	fputs("ERROR: No pages found!\n", stderr);
	return (1);
    } else {
	fputs("INFO: Ready to print.\n", stderr);
	return (0);
    }
}


